from multiprocessing import Pool
import os
from Bio import Phylo
import pandas as pd
from tqdm import tqdm

# list of functions to make
# break tree into pruned tree and subtrees
def DivideTree(treefile, clusterfile):
    '''
    Take a tree and condense based on clustering definitions, into a main tree with just the clusters as leaves
    and all the different subtrees that are children of the main tree.

    Args:
    - treefile (str): path to tree file
    - clusters (str): path to cluster file
    '''

    # read in tree
    tree = Phylo.read(treefile, 'newick')
    clusters=pd.read_csv(clusterfile, dtype=str)

    # get all internal nodes that represent clusters
    anc_data=[data['id'].to_list() for cluster, data in clusters.groupby('cluster') if len(data)>1]
    cluster_order=[cluster for cluster, data in clusters.groupby('cluster') if len(data)>1]
    # do multi-processing to find common_ancestor
    with Pool() as pool:
        results=pool.map(tree.common_ancestor, anc_data)
    anc_nodes=dict(zip(cluster_order, results))

    # split all the subtrees
    subtrees={}
    for cluster, anc_node in anc_nodes.items():
        subtrees[cluster]=Phylo.BaseTree.Tree(root=anc_node)
        # make root node distance 0
        subtrees[cluster].root.branch_length=0
    
    # create the main tree where nodes are collapsed into leaves if they represent clusters
    target_clades=[clade.name for clade in anc_nodes.values()]#+leaves

    PruneTree(tree, target_clades)

    return tree, subtrees, anc_nodes

# Function to prune the tree
def PruneTree(tree, target_clades):
    """
    Prune the tree by removing branches and making internal nodes as leaves.

    Parameters:
    - tree (Phylo.BaseTree.Tree): The tree to prune.
    - target_clades (list of str): List of internal clade names to retain
    """
    # get depths 
    depths = tree.depths()

    # Then, convert internal nodes to leaves if they are in the target_clades
    for clade in tree.get_nonterminals(order="postorder"):
        if clade.name in target_clades:
            # get terminals to determine deepest leaf
            leaf_depths=[depths[leaf] for leaf in clade.get_terminals()]
            depth_diff=max(leaf_depths)-depths[clade]
            clade.branch_length+=depth_diff
            # remove children
            clade.clades = []

# functions to generate ITOL files
def ITOLMultiCategoryHeader(dataset_label, field_colors, field_labels, datatype='piechart'):
    header = (
        f"DATASET_{datatype.upper()}\n"
        # "#In pie chart datasets, each ID is associated to multiple numeric values...\n"
        "SEPARATOR COMMA\n"
        f"DATASET_LABEL,{dataset_label}\n"
        # "COLOR,#ff0000\n"
        f"FIELD_COLORS,{','.join(field_colors)}\n"
        f"FIELD_LABELS,{','.join(field_labels)}\n"
        "#=================================================================#\n"
        "DATA\n"
    )
    return header

def ITOLProportionDataset(annot_table, color_annot, other_color='#bdbdbd', datatype='multibar', percentage=False, outdir=''):
    '''
    Make a dataset file for ITOL to display proportions of categorical data.

    Args:
    - annot_table (pd.DataFrame): table with columns 'id', 'value', 'count'. Can be obtained from 'AnnotateClusters' function (select a level by key).
    - color_annot (dict): dictionary with keys 'categories', 'value', 'method'. Can be obtained from 'ColorAnnot' function.
    - other_color (str): color to use for the 'Other' category. Aggregates all data not in the color_annot
    - datatype (str): type of dataset to make. Either 'multibar' or 'piechart'
    - percentage (bool): whether to display the data as percentage
    - outdir (str): directory to save the file
    '''
    if datatype not in ['multibar', 'piechart']:
        raise ValueError(f"datatype must be either 'multibar' or 'piechart', not {datatype}")
    
    # go through the data set and 
    color_dict=color_annot['categories']
    label = f'{color_annot['value']}_{color_annot['method']}'.capitalize()
    main_colors, main_labels = list(color_dict.values()), list(color_dict.keys())
    field_colors, field_labels = main_colors + [other_color], main_labels + ['Other']
    # more dataset specific 
    extra_fields = '-1,10,' if datatype == 'piechart' else ''
    suffix = '_norm' if percentage else ''

    # iterate through data and create annotation
    with open(os.path.join(outdir, f"{label}_{datatype}_dataset{suffix}.txt"), "w") as file:
        header = ITOLMultiCategoryHeader(f"{label} {datatype}", field_colors, field_labels, datatype)
        file.write(header)
        for id, data in annot_table.groupby('id'):
            # aggregate all data not in field colors as other
            main_colored=data[data['value'].isin(main_labels)]
            other_count=data[~data['value'].isin(main_labels)]['count'].sum()
            data_dict=main_colored.set_index('value')['count'].to_dict()
            data_dict['Other']=other_count
            counts=[data_dict.get(label, 0) for label in field_labels]
            if percentage:
                total=sum(counts)
                counts=[count/total*100 for count in counts]
            data_line = f"{id},{extra_fields}{','.join(map(str, counts))}\n"
            file.write(data_line)

def ITOLColoredRangesDataset(dataset_label, branch_values, color_dict, outdir='', color_other=True, other_color='#bdbdbd'):
    '''
    Make a colored branches dataset file for ITOL

    Args:
    - dataset_label (str): label for the dataset
    - branch_values (dict): dictionary with branch id as key and value as value
    - color_dict (dict): dictionary with value as key and color as value. Obtained from 'ColorAnnot' function (key='categories')
    - outdir (str): directory to save the file
    - color_other (bool): whether to color branches not in the color_dict as 'Other'
    - other_color (str): color to use for the 'Other' category. Aggregates all data not in the color_annot
    '''
    legend_dict={value: color for value, color in color_dict.items()}
    if color_other:
        legend_dict['Other']=other_color
    with open(os.path.join(outdir, f"{dataset_label}_colored_ranges_dataset.txt"), "w") as file:
        header = (
            f"DATASET_RANGE\n"
            "SEPARATOR COMMA\n"
            f"DATASET_LABEL,{dataset_label}\n"
            "COVER_DATASETS,1\n"
            f"LEGEND_COLORS,{','.join(legend_dict.values())}\n"
            f"LEGEND_LABELS,{','.join(legend_dict.keys())}\n"
            f"LEGEND_TITLE,{dataset_label}\n"
            f"LEGEND_SHAPES,{','.join(['1']*len(legend_dict))}\n"
            "#=================================================================#\n"
            "DATA\n"
        )
        file.write(header)
        for branch, value in branch_values.items():
            color=color_dict.get(value, other_color if color_other else None)
            if color is None:
                continue
            # data_line = f"{branch},label,clade,#000000,1,normal,{color}\n"
            data_line=f'{branch},{branch},{color},,,,,,,,,\n'
            file.write(data_line)


def ITOLBinaryDataset(dataset_label, leaf_ids, shape=2, color='#ff0000', outdir='', internal=False):
    '''
    Make a binary dataset file for ITOL

    Args:
    - dataset_label (str): label for the dataset
    - leaf_ids (list of str): list of leaf ids in the tree to highlight
    - shape (int): shape to show in ITOL, 
        -    #1: square
        -    #2: circle
        -    #3: star
        -    #4: right pointing triangle
        -    #5: left pointing triangle
        -    #6: checkmark
    - color (str): color of the markers
    - outdir (str): directory to save the file
    - internal (bool): whether to make an internal node dataset
    '''
    datatype='binary' if not internal else 'symbol'
    outlabel='leaf' if not internal else 'internal'
    with open(os.path.join(outdir, f"{dataset_label}_highlight_{outlabel}_dataset.txt"), "w") as file:
        header_front = (
            f"DATASET_{datatype.upper()}\n"
            "SEPARATOR COMMA\n"
            f"DATASET_LABEL,{dataset_label}\n")
        header_middle = (
            f"COLOR,{color}\n"
            f"FIELD_SHAPES,{shape}\n"
            f"FIELD_LABELS,{dataset_label}\n") if not internal else ("MAXIMUM_SIZE,5\n")
        header_end = (
            "#=================================================================#\n"
            "DATA\n"
        )
        header=header_front+header_middle+header_end
        file.write(header)

        for id in leaf_ids:
            if not internal:
                data_line = f"{id},1\n"
            else:
                data_line = f"{id},{shape},1,{color},1,1\n"
            file.write(data_line)

def ITOLLabelDataset(dataset_label, label_dict, outdir):
    '''
    Make a label dataset file for ITOL

    Args:
    - dataset_label (str): label for the dataset
    - label_dict (dict): dictionary with id as key and label as value
    - outdir (str): directory to save the file
    '''
    with open(os.path.join(outdir, f"{dataset_label}_label_dataset.txt"), "w") as file:
        header = (
            f"DATASET_TEXT\n"
            "SEPARATOR COMMA\n"
            f"DATASET_LABEL,{dataset_label}\n"
            # "COLOR,#000000\n"
             "#=================================================================#\n"
            "DATA\n"
        )
        file.write(header)
        for id, label in label_dict.items():
            data_line = f"{id},{label},-1,#000000,normal,1,0\n"
            file.write(data_line)
        
    


import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import os

def read_and_sample(file_path, column_name, max_points=1000000, sep='\t'):
    # Estimate the size of the file
    file_size = os.path.getsize(file_path)

    # Assuming each point in the column takes approximately 8 bytes (for float)
    estimated_points = file_size / 8

    # If the estimated number of points is less than max_points, read the entire column
    if estimated_points <= max_points:
        data = pd.read_csv(file_path, usecols=[column_name], sep=sep)[column_name]
    else:
        # Sample the points
        skip = estimated_points // max_points
        data = pd.read_csv(file_path, usecols=[column_name], skiprows=lambda i: i % skip != 0, sep=sep)[column_name]

    return data

def calculate_percentiles(data, target_percentiles):
    percentiles = data.quantile(target_percentiles)
    return dict(zip(target_percentiles, percentiles))

def plot_distribution(data, column_name, plot_out=None, show=True):
    sns.kdeplot(data, fill=True)
    plt.xlabel(column_name)
    plt.ylabel('Frequency')
    if plot_out is not None:
        plt.savefig(plot_out)
    if show:
        plt.show()

def Summarize(file_path, column_name, plot_out=None, target_percentiles=[.1, .25, .5, .75, .9]):
    data = read_and_sample(file_path, column_name)
    percentiles = calculate_percentiles(data, target_percentiles)
    plot_distribution(data, column_name, plot_out=plot_out, show=False)
    return {'percentiles':percentiles, 'plot':plot_out}




from .tools.cluster_merging import SetsToDataFrame, iterative_merge
from .tools.cluster_stream import ClusterStreamMulti
from .tools.hierarchical_clustering import FlattenHierarchy, HierarchicalClustering, linkage_to_newick, NameInternalNodes
from .tools.mmseqs_wrapper import RunMMseqs
from .tools.summarize_distribution import read_and_sample, calculate_percentiles, plot_distribution
from .tools.esm_wrapper import Embed
from .tools.sanitize_fasta_headers import sanitize_fasta_headers
from .tools.kmeans import KMeansBisecting, BisectingKMeansKeepTree, compute_linkage_from_tree, tree_to_newick

import matplotlib.pyplot as plt
import pickle
import pandas as pd
import numpy as np
import os
from sklearn.decomposition import PCA
from umap import UMAP
import shutil
from scipy.cluster.hierarchy import fcluster
from Bio import Phylo

def StateFile(out_label):
    return f'{out_label}_states.pkl'

def SaveStates(states, state_file):
    # read in state file
    old_states=pickle.load(open(state_file, 'rb')) if os.path.exists(state_file) else {}
    # update states
    old_states.update(states)
    with open(state_file, 'wb') as f:
        pickle.dump(old_states, f)

def PathToFile(directory, path):
    return os.path.join(directory, path)

def LoadEmbeddings(directory, states, filter_path=None, custom_embeddings=None):
    embeddings_file= custom_embeddings if custom_embeddings else PathToFile(directory, states['embeddings'])
    embeddings=pickle.load(open(embeddings_file, 'rb'))
    if filter_path:
        # read in filter as a set
        print('Filtering embeddings by id')
        filter_set=set()
        with open(filter_path, 'r') as f:
            for line in f:
                filter_set.add(line.strip())
        # filter embeddings
        embeddings={k:v for k,v in embeddings.items() if k in filter_set}
    # convert embeddings from pytorch tensor to numpy array
    ids=list(embeddings.keys())
    embeddings=np.array(list(embeddings.values()))

    return ids, embeddings

def ReduceVectors(embeddings, keep_features):
    # get length of 1st vector
    vector_length=len(embeddings[0])

    if keep_features < vector_length and keep_features > 0:
        print('Reducing feature dimensions with PCA')
        pca = PCA(n_components=keep_features)
        return pca.fit_transform(embeddings)
    else:
        return embeddings

import argparse

parser = argparse.ArgumentParser(description='Pipeline for analyzing protein families using unsupervised clustering. Uses either homology or vector embeddings to cluster sequences.',
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)

ga=parser.add_argument_group('General arguments')
ga.add_argument('-d', '-directory', help='Output directory for files', default='out/')
ga.add_argument('-p', '-prefix', help='Prefix for output files', default='ssn')
ga.add_argument('-fa', '-fasta', help='Fasta file to analyze. Only needs to be given once.', default=None)

hom=parser.add_argument_group('Homology based method')
# mmseqs all-by-all
hom.add_argument('-A','-all_by_all', action='store_true', help='Run mmseqs all-by-all search')
# mmseqs all-by-all clustering
hom.add_argument('-MC','-mmseqs_cluster', action='store_true', help='Cluster mmseqs results')
hom.add_argument('-P', '-cluster_percentiles', action='store_true', help='Use percentiles of mmseqs results to do clustering.')
hom.add_argument('-cc', '-cluster_cutoffs', nargs='+', type=float, help='Manually set cutoffs for clustering. Is ignored if -P is set.')
hom.add_argument('-cluster_lines', type=int, help='Number of lines to read at once, adjust for memory consumption.', default=10_000_000)
hom.add_argument('-f', '-filter', help='List of IDs to filter for. Can be used for mmseqs clustering, kmeans, and UMAP.')
hom.add_argument('-RMC', '-redo_mmseqs_clustering', action='store_true', help='Redo steps for mmseqs clustering.')
hom.add_argument('-cluster_jobs', type=int, help='Number of jobs to use for clustering', default=None)
hom.add_argument('-edge_table', help='Path to user provided edge table for clustering. Needs to be a tab-separated file with \'target\'/\'query\' in columns 1 and 2, and the distance score in column 3 (header names do not matter)', default=None)

# esm embedding
emb=parser.add_argument_group('Vector embedding')
emb.add_argument('-E','-embed_vectors', action='store_true', help='Embed vectors')
emb.add_argument('-CE','-continue_embed', action='store_true', help='Continue embedding vectors, in case of interrupted run.')
emb.add_argument('-t', '-tok_per_batch', type=int, help='Number of tokens per batch for embedding. Reduce if running out of memory.', default=30_000)

vec=parser.add_argument_group('Vector based methods')
# general vector
vec.add_argument('-KF', '-keep_features', type=int, help='Number of features to use in vector based methods.', default=30)
vec.add_argument('-embeddings_dict', help='Path to user provided embeddings dictionary. Needs to be a pickle file with a dictionary of embeddings, where the keys are the sequence IDs and the values are the embeddings vectors (1D, already pooled).', default=None)
# Kmeans clustering
vec.add_argument('-KM', '-kmeans', action='store_true', help='Run kmeans clustering')
vec.add_argument('-K', '-kmeans_clusters', nargs='+', type=int, help='Number of kmeans clusters. Also used for setting cluster count when flattening hierachical clusters.', default=[])
vec.add_argument('-FKM', '-flatten_kmeans', action='store_true', help='Subsect kmeans clusters based on max_k tree')
vec.add_argument('-max_k', type=int, help='Max K to further subdivide into clusters')
# Hierarchical clustering
vec.add_argument('-HC', '-hierarchical_clustering', action='store_true', help='Run hierarchical clustering')
vec.add_argument('-CT', '-convert_hierarchical_tree', action='store_true', help='Convert hierarchical tree to newick format')
# Flatten hierarchical clustering
vec.add_argument('-FHC', '-flatten_hierarchical_clustering', action='store_true', help='Flatten hierarchical clustering')
# UMAP
vec.add_argument('-U', '-umap', action='store_true', help='Run UMAP')
args=parser.parse_args()

##### Settings to human readable #####
# This is where the output data will go, includes the prefix for the output files, and the output directory
out_directory=args.d
out_prefix=args.p

all_by_all=False or args.A # run mmseqs all-by-all search

embed_vectors=False or args.E # embed vectors
continue_embed=False or args.CE # continue embedding vectors, use if previous run was somehow interrupted
tok_per_batch=args.t # number of tokens per batch for embedding

mmseqs_cluster=False or args.MC # cluster mmseqs results
use_percentiles=False or args.P # use percentiles of mmseqs results to do clustering
cluster_cutoffs=args.cc # cutoffs for clustering
custom_edge_table=args.edge_table # path to user provided edge table for clustering

custom_embeddings=args.embeddings_dict # path to user provided embeddings dictionary
run_kmeans=False or args.KM # run kmeans clustering
kmeans_clusters=args.K # number of kmeans clusters
keep_features=args.KF # number of features to use in kmeans clustering
flatten_kmeans=False or args.FKM # subsect kmeans clusters based on max_k tree
flatten_kmeans_max_k=args.max_k # max K to further subdivide into clusters

run_hierarchical_clustering=False or args.HC # run hierarchical clustering
convert_hierarchical_tree=False or args.CT # convert hierarchical tree to newick format
flatten_hierarchical_clustering=False or args.FHC # flatten hierarchical clustering

run_umap=False or args.U # run UMAP
####################

# make sure the output directory exists
os.makedirs(out_directory, exist_ok=True)

# make a file to keep track of outputs
state_file=StateFile(PathToFile(out_directory, out_prefix))
if os.path.exists(state_file):
    with open(state_file, 'rb') as f:
        states=pickle.load(f)
else:
    states={}

# the fasta file to be analyzed
fasta=args.fa

if 'cleaned_fasta' not in states:
    # clean up fasta headers by converting them to an index
    fasta_info=sanitize_fasta_headers(fasta, out_directory, out_prefix)

    # save the state
    print(f'files produced in {out_directory}:', fasta_info)
    states.update(fasta_info)
    SaveStates(states, state_file)

# make the path to the file, for other analyses
cleaned_fasta=PathToFile(out_directory, states['cleaned_fasta'])
print('cleaned fasta is in:', cleaned_fasta)

# Run mmseqs
if all_by_all:
    RunMMseqs_out=RunMMseqs(cleaned_fasta, out_directory, out_prefix)

    print('Updates to state:', RunMMseqs_out)
    states.update(RunMMseqs_out)
    SaveStates(states, state_file)

if 'mmseqs_result' in states and 'mmseqs_percentiles' not in states:
    # # define paths to data
    mmseqs_result=PathToFile(out_directory, states['mmseqs_result'])
    image_out=PathToFile(out_directory, f'{out_prefix}_bits.png')
    target_metric='bits' # metric stored in the mmseqs_result table

    # helper functions that finds the 10, 25, 50, 75, and 90 percentiles of the pairwise edge table, and plots the distribution of the data
    data = read_and_sample(mmseqs_result, 'bits', max_points=20_000_000) # subsample if there are more than X entries
    percentiles = calculate_percentiles(data, [.1, .25, .5, .75, .9]) # provide target precentiles as fractions
    print('percentiles:', percentiles)
    plot_distribution(data, 'bits', plot_out=image_out, show=False)

    states['mmseqs_percentiles']=percentiles
    SaveStates(states, state_file)

# cluster mmseqs results
if mmseqs_cluster:

    if use_percentiles:
        cutoffs=list(states['mmseqs_percentiles'].values())
    else:
        cutoffs=cluster_cutoffs

    if 'mmseqs_cluster_cutoffs' not in states:
        states['mmseqs_cluster_cutoffs']={}
    for cutoff in cutoffs:
        if cutoff not in states['mmseqs_cluster_cutoffs']:
            states['mmseqs_cluster_cutoffs'][cutoff]='started'
    SaveStates(states, state_file)

    # cluster the mmseqs results in chunks
    edge_table_path= custom_edge_table if custom_edge_table else PathToFile(out_directory, states['mmseqs_result'])
    cluster_path=os.path.join(out_directory, 'mmseqs_clustering/')
    if not os.path.exists(cluster_path):
        os.makedirs(cluster_path)

    use_cutoffs=[x for x in cutoffs if x not in states['mmseqs_cluster_cutoffs'] or states['mmseqs_cluster_cutoffs'][x] == 'started'] if not args.RMC else cutoffs
    if len(use_cutoffs) == 0:
        print('All mmseqs clusterings are already done. Proceeding to cluster merging.')
    else:
        print('Clustering mmseqs results with cutoffs:', use_cutoffs)
        ClusterStreamMulti(
            edge_table_path, 
            use_cutoffs, 
            query=0, target=1, score=2, 
            has_header=True, field_separator="\t", 
            # label=out_prefix, 
            directory=cluster_path, 
            chunksize=args.cluster_lines, n_jobs=None, filter_path=args.f)
    
    for cutoff in use_cutoffs:
        if states['mmseqs_cluster_cutoffs'][cutoff] == 'started':
            states['mmseqs_cluster_cutoffs'][cutoff]='chunked'
    SaveStates(states, state_file)

    # merge the chunked clusterings into a single file
    merge_cutoffs=[x for x in cutoffs if (states['mmseqs_cluster_cutoffs'][x] == 'chunked' or args.RMC)]
    print('Merging mmseqs clusters:', merge_cutoffs)
    for cutoff in merge_cutoffs:
        folder=str(cutoff)
        print('Processing folder:',folder)
        folder_path=os.path.join(cluster_path, folder)
        files = [x for x in os.listdir(folder_path) if x.endswith('.pkl')]
        datasets = [os.path.join(folder_path, x) for x in files]
        merged_dataset = iterative_merge(datasets, max_datasets=args.cluster_jobs)
        # save the merged dataset
        out_file = os.path.join(cluster_path, f'{out_prefix}_mmseqs_clusters_{cutoff}.csv')
        SetsToDataFrame(merged_dataset).to_csv(out_file, index=False)
        print('Saved:', out_file)
        if states['mmseqs_cluster_cutoffs'][cutoff] == 'chunked':
            states['mmseqs_cluster_cutoffs'][cutoff]='done'
        SaveStates(states, state_file)

        # remove the folder with the chunked clusterings
        shutil.rmtree(folder_path)

if embed_vectors:
    embed_out=Embed(cleaned_fasta, out_directory, out_prefix, tok_per_batch=tok_per_batch, cont_run=continue_embed) # change tok_per_batch lower if you run out of GPU memory

    print('Updates to state:', embed_out)
    states.update(embed_out)
    SaveStates(states, state_file)

if run_kmeans:
    # kmeans_path=os.path.join(out_directory, 'kmeans/max_k/')
    # # make sure the output directory exists
    # os.makedirs(kmeans_path, exist_ok=True)

    print('Loading embeddings for kmeans')
    ids, embeddings=LoadEmbeddings(out_directory, states, args.f, custom_embeddings)

    # get length of 1st vector
    embeddings=ReduceVectors(embeddings, keep_features)

    # run kmeans
    print('Running kmeans')
    # KMeansBisecting(embeddings, kmeans_clusters, ids, kmeans_path, out_prefix)

    # save each cluster file as a 'max K' file, that is then fed to a lower K clustering
    new_states={}
    for k in kmeans_clusters:
        kmeans = BisectingKMeansKeepTree(n_clusters=k, random_state=0)
        kmeans.fit(embeddings)

        # paths
        label_file=f'kmeans/max_k/k{k}_labels.csv'
        tree_file=f'kmeans/max_k/k{k}_tree.nwk'
        os.makedirs(os.path.join(out_directory, 'kmeans/max_k/'), exist_ok=True)


        labels=pd.DataFrame({'id':ids, 'label':kmeans.labels_+1})
        labels.to_csv(os.path.join(out_directory, label_file), index=False)
        # make tree structure
        tree=tree_to_newick(kmeans._bisecting_tree)

        # write tree to file
        with open(os.path.join(out_directory, tree_file), 'w') as f:
            f.write(tree)
        
        new_states[k]={'labels':label_file, 'tree':tree_file}
    states['max_k']=new_states
    SaveStates(states, state_file)

if flatten_kmeans:
    # check if max_k exists
    if flatten_kmeans_max_k is None:
        print('Please provide a max K to subsect kmeans clusters. (use -max_k)')
    elif flatten_kmeans_max_k not in states['max_k']:
        print(f'Max K not found in states. There are {len(states["max_k"])} max K values processed: {list(states["max_k"].keys())}')
 
    # read tree
    tree_file=PathToFile(out_directory, states['max_k'][flatten_kmeans_max_k]['tree'])
    tree=Phylo.read(tree_file, 'newick')
    linkage_matrix, leaf_names = compute_linkage_from_tree(tree)
    labels_file=PathToFile(out_directory, states['max_k'][flatten_kmeans_max_k]['labels'])
    labels=pd.read_csv(labels_file)

    # make sure the output directory exists
    k_out_path=os.path.join(out_directory, f'kmeans/flattened_{flatten_kmeans_max_k}/')
    # print(k_out_path)
    os.makedirs(k_out_path, exist_ok=True)

    # create smaller clusters
    for smaller_k in kmeans_clusters:
        if smaller_k > flatten_kmeans_max_k:
            continue
        print('Flattening kmeans clusters to:', smaller_k)
        # Apply fcluster to get cluster labels
        cluster_labels = fcluster(linkage_matrix, smaller_k, criterion='maxclust')

        # write labels to file
        leaf_labels=pd.DataFrame({'label':leaf_names, 'cluster':cluster_labels})

        # merge on labels
        merged=labels.merge(leaf_labels, on='label')
        # write to file
        # print(os.path.join(k_out_path, f'k{smaller_k}_clusters.csv'))
        merged[['id','cluster']].to_csv(os.path.join(k_out_path, f'{out_prefix}_kmeans_bisecting_{smaller_k}.csv'), index=False)

if run_umap:
    # load vector embeddings
    print('Loading embeddings for UMAP')
    ids, embeddings=LoadEmbeddings(out_directory, states, args.f, custom_embeddings)

    embeddings=ReduceVectors(embeddings, keep_features)
    
    # perform UMAP
    print('Performing UMAP')
    umap=UMAP(n_components=2)
    umap_embeddings=umap.fit_transform(embeddings)

    # make dataframe
    umap_df=pd.DataFrame(umap_embeddings, columns=['x', 'y'])
    # add sequence ids
    umap_df['id']=ids

    # save and save to state
    umap_file=out_prefix+'_umap.csv'
    umap_df.to_csv(PathToFile(out_directory, umap_file), index=False)
    states['umap']=umap_file
    SaveStates(states, state_file)

if run_hierarchical_clustering:
   
    ids, embeddings=LoadEmbeddings(out_directory, states, args.f, custom_embeddings)

    embeddings=ReduceVectors(embeddings, keep_features)

    # run hierarchical clustering
    hc_out=HierarchicalClustering(ids, embeddings, out_directory, out_prefix)
    print('Updates to state:', hc_out)
    states.update(hc_out)

    SaveStates(states, state_file)

if convert_hierarchical_tree and 'linkage_matrix' in states:
    # see if distance matrix exists
    matrix=pickle.load(open(PathToFile(out_directory, states['linkage_matrix']), 'rb'))
    ids=pickle.load(open(PathToFile(out_directory, states['linkage_ids']), 'rb'))

    # convert linkage matrix to newick format
    newick=linkage_to_newick(matrix, ids)
    # save newick format
    outfile=f'{out_prefix}_hc_tree.nwk'
    outpath=os.path.join(out_directory, outfile)
    with open(outpath, 'w') as f:
        f.write(newick)
    
    # rename internal nodes
    NameInternalNodes(outpath)

    states.update({'hc_tree':outfile})
    SaveStates(states, state_file)

if flatten_hierarchical_clustering:
    hc_path=os.path.join(out_directory, 'hierarchical_clustering/')
    # make sure the output directory exists
    os.makedirs(hc_path, exist_ok=True)

    # load linkage matrix
    linkage_matrix=PathToFile(out_directory, states['linkage_matrix'])
    linkage_ids=PathToFile(out_directory, states['linkage_ids'])
    cutoffs=kmeans_clusters

    # flatten hierarchy
    FlattenHierarchy(linkage_matrix, linkage_ids, cutoffs, hc_path, out_prefix)


import pandas as pd

def AnnotateClusters(clusters, levels, annot_table, column, cluster_key, annot_key,  numeric_func='mean', dropna=False):
    '''
    Function to aggregate annotations at the cluster level.

    Arguments:
    - clusters: DataFrame with cluster assignments, with each level of clustering as its own column
    - levels: List of levels (columns) to aggregate annotations within 'clusters'
    - annot_table: DataFrame with annotations to aggregate
    - column: Column in 'annot_table' to aggregate
    - cluster_key: Column in 'clusters' that contains the sequence ID
    - annot_key: Column in 'annot_table' that contains the sequence ID
    - numeric_func: Function to aggregate numeric data, default is 'mean'. Can be any string or function accepted by the Pandas.groupby.agg function.
    - dropna: Whether to drop NA values when aggregating numeric data. Default is False.

    Returns a dictionary with the following keys
    - method: The method used to aggregate the data
    - label: The label of the annotation
    - levels: List of levels
    - value: The column in 'annot_table' that was aggregated, useful reference for labelling legends
    - For each level in 'levels', a dictionary with the following keys:
        - id: The cluster ID
        - value: The aggregated value    
    '''

    is_numeric=pd.api.types.is_numeric_dtype(annot_table[column])
    method=numeric_func if is_numeric else 'counts'
    out={'method':method, 'label':f'{column}_{method}', 'levels':levels, 'value':column}
    for level in levels:
        sel=clusters[[cluster_key,level]].copy()
        data=sel.merge(annot_table[[annot_key, column]], left_on=cluster_key, right_on=annot_key)
        annot=data.groupby(level)
        
        if method!='counts':
            annot=annot[column].agg(numeric_func)
        else:
            annot=annot[column].value_counts(dropna=dropna)

        out[level]=annot.reset_index().rename(columns={level:'id', column:'value'})
    
    return out

import colorsys

def ColorSaturation(rgba, saturation):
    """
    Desaturate an RGB color.

    :param rgb: A tuple of (R, G, B) in the range 0-255.
    :param desaturation_factor: Float between 0 (no change) and 1 (fully desaturated).
    :return: Desaturated RGB color, as a tuple in the range 0-255.
    """
    # Normalize RGB values to 0-1 range
    r, g, b = [x  for x in rgba[:3]]

    # Convert RGB to HSV
    h, s, v = colorsys.rgb_to_hsv(r, g, b)

    # Reduce the saturation
    s = saturation

    # Convert back to RGB
    desaturated_rgb = colorsys.hsv_to_rgb(h, s, v)

    # Convert back to 0-255 range
    return [x for x in desaturated_rgb]+[rgba[3]]

import matplotlib as mpl

def AdjustColormapSaturation(cmap, saturation=1):
    """
    Adjust the saturation of a Matplotlib colormap.

    :param cmap: Original Matplotlib colormap.
    :param saturation_factor: Saturation factor (0-1, where 1 is full saturation).
    :return: New colormap with adjusted saturation.
    """
    # Create a new colormap array based on the original colormap
    new_cmap_data = cmap(range(cmap.N))

    # Adjust the saturation for each color in the colormap
    for i, rgba in enumerate(new_cmap_data):
        # Convert RGBA to HSV
        h, s, v = colorsys.rgb_to_hsv(rgba[0], rgba[1], rgba[2])
        # Adjust saturation
        s *= saturation
        # Convert back to RGB
        r, g, b = colorsys.hsv_to_rgb(h, s, v)
        # Update the new colormap array
        new_cmap_data[i] = r, g, b, rgba[3]

    # Create a new colormap from the adjusted array
    return mpl.colors.ListedColormap(new_cmap_data)

def AssignCategoricalColors(all_cats, color_map, shuffle_colors_seed=None, is_categorical=True, top_n=None):

    include=all_cats if top_n is None else all_cats[:top_n]

    # make a color map, give each category a color
    c_mapping={}
    for i, cat in enumerate(include):
        if is_categorical:
            c_mapping[cat]=color_map(i%color_map.N)
        else:
            c_mapping[cat]=color_map(i/(len(include)-1))
    
    if shuffle_colors_seed is not None:
        # seed random number generator
        Random=random.Random(shuffle_colors_seed)
        # shuffle the colors
        k, v=[list(x) for x in zip(*c_mapping.items())]
        Random.shuffle(v)
        c_mapping=dict(zip(k,v))
    
    return c_mapping
        
import random
c_template='rgba({},{},{},{})'
def FormatColor(color, format='bokeh', scale=255):
    '''
    Args:
        color: Iterable of RGBA values in the range 0-1.
        format: 'bokeh', 'hex' or 'tuple'
        scale: Scale to multiply the color values by. Default is 255, assuming input is 0-1.
    '''
    color=[int(x*scale) for x in color]
    if format=='bokeh':
        return c_template.format(*color)
    elif format=='hex':
        return "#{:02x}{:02x}{:02x}".format(*color[:3])
    elif format=='tuple':
        return tuple(color)

import multiprocessing as mp

def ColorAnnot( annot:dict, cmap='viridis', top_n=None, saturation=1, shuffle_colors_seed=None, 
                vmin=None, vmax=None, cmap_is_categorical=True, binary_blend=False, color_format='bokeh', direct_map_colors=False, annot_order=None, min_count=0, levels=None):
    '''
    Colors a set of annotations made with AnnotateClusters.

    Arguments:
    - annot: Dictionary with annotations, as returned by AnnotateClusters
    - cmap: Colormap to use. Can be a string (name of a Matplotlib colormap) or a Matplotlib colormap object.
    - top_n: Number of top categories to color. Default is None, which colors all categories.
    - saturation: Saturation of the colormap. Default is 1.
    - shuffle_colors_seed: Seed for shuffling colors in categorical data. Mostly useful when there are many categories and there is no preference for color order.
                         Default is None, which does not shuffle colors. 
    - vmin: Minimum value for a numeric colormap. Default is None, which uses the minimum value in the data.
    - vmax: Maximum value for a numeric colormap. Default is None, which uses the maximum value in the data.
    - cmap_is_categorical: When a colormap object is given (instead of a string), if the colormap should be considered categorical. Only applies for categorical data. Default is True.
    - binary_blend: For blending colors. If True, treat all instances in categorical data to be yes/no (i.e., max count of 1 if present). Default is False.
    - color_format: Format of the color to output. Can be 'bokeh', 'hex' or 'tuple'. Default is 'bokeh'.
    - direct_map_colors: If True, map the color directly to the category (assumes only one color per id). Default is False, which blends the colors based on the count of each category.
    - annot_order: Order of annotations for categorical data. Default is None, which orders by overall count in the dataset.

    Returns a dictionary with the following:
    - value: The column in 'annot_table' that was aggregated, useful reference for labelling legends
    - method: The method used to aggregate the data
    - min: Minimum value in the data
    - max: Maximum value in the data
    - colorscale: List of colors in the colormap
    - categories: Dictionary with categories as keys and colors as values. Only present if the data is categorical.
    - For each level in 'levels', a dictionary with the following keys:
        - id: The cluster ID
        - value: The color of the cluster
        
    '''

    # settle color scheme
    if type(cmap)==str:
        color_map=mpl.colormaps.get_cmap(cmap)
        is_categorical=cmap in ['tab10', 'tab20', 'tab20b', 'tab20c','Pastel1', 'Pastel2', 'Paired', 'Set1', 'Set2', 'Set3','Accent','Dark2']
    else:
        color_map=cmap
        is_categorical=cmap_is_categorical
    color_map=AdjustColormapSaturation(color_map, saturation)
    levels=annot['levels'] if levels is None else levels

    colors={'value':annot['value'], 'method':annot['method']}
    if annot['method']!='counts':
        # determine range of values
        # Normalize data (scale it to range between 0 and 1)
        min_value=min([min(annot[x]['value']) for x in annot['levels']]) if vmin is None else vmin
        max_value=max([max(annot[x]['value']) for x in annot['levels']]) if vmax is None else vmax

        # Choose a color map
        cmap_resolution=100
        colors['colorscale'] = [[i/(cmap_resolution-1), FormatColor(color_map(i/(cmap_resolution-1)), color_format)] for i in range(cmap_resolution)]
        colors['min']=min_value
        colors['max']=max_value
        
        for level in levels:
            data=annot[level].copy()
            values=data['value']
            normalized_data = (values - min_value) / (max_value - min_value)

            # Map the normalized data to colors from the color map
            mapped_colors = color_map(normalized_data)

            # format rgba string
            mapped_colors=[FormatColor(c, color_format) for c in mapped_colors]

            # save colors alongside with id
            colors[level]=dict(zip(data['id'], mapped_colors))
    else:
        # get all categories
        
        all_levels=[]
        for level in levels:
            all_levels.append(annot[level])
        
        if annot_order is not None:
            all_cats=annot_order
        else:
            all_level_df=pd.concat(all_levels).dropna(subset=['value']) # ignore NA values when distributing colors
            # all categories in order of count
            all_cats=all_level_df.groupby('value')['count'].sum().reset_index().sort_values('count', ascending=False)['value'].values
        c_mapping=AssignCategoricalColors(all_cats, color_map, shuffle_colors_seed, is_categorical, top_n)
        colors['categories']={k:FormatColor(v, color_format) for k,v in c_mapping.items()}
        
        for level in levels:
            data=annot[level].copy()
            # map the category to the color
            data['color']=data['value'].map(c_mapping)
            # blend the colors by weight
            blended={}
            if direct_map_colors:
                data.dropna(subset=['color'], inplace=True)
                colors[level]=dict(zip(data['id'], data['color'].apply(lambda x: FormatColor(x, color_format))))
                continue
            # use multiprocessing to speed up the process

            inputs=[]
            ids=[]
            for id, subdata in data.groupby('id'):
                if min_count>0:
                    subdata=subdata[subdata['count']>=min_count]
                inputs.append(subdata)
                ids.append(id)

            with mp.Pool() as pool:
                # blended=dict(zip(data['id'], pool.starmap(BlendedClusterColor, [(subdata, c_mapping, binary_blend, color_format) for _, subdata in data.groupby('id')])))
                results=pool.starmap(BlendedClusterColor, [(subdata, c_mapping, binary_blend, color_format) for subdata in inputs])
            blended=dict(zip(ids, results))
            colors[level]=blended
        
    return colors

def BlendedClusterColor(subdata, c_mapping, binary_blend=False, color_format='bokeh'):
    r,g,b=0,0,0
    total=subdata['count'].sum() if not binary_blend else len(subdata)
    for row in subdata.itertuples():
        weight=row.count if not binary_blend else 1
        if row.value not in c_mapping: 
            r+=weight
            g+=weight
            b+=weight
        else:
            r+=row.color[0]*weight
            g+=row.color[1]*weight
            b+=row.color[2]*weight
        
    if total==0:
        return 'rgba(0,0,0,0)'
    else:
        r/=total
        g/=total
        b/=total
        return FormatColor((r,g,b,1), color_format)


### Unused below
def OutlineClusters(clusters, levels, targets, rep_map=None, mapping:dict=None, map_func=None, color='red', label='selected', alpha=.7):
    '''
    rep_map: Dataframe with 'rep' and 'id' columns. Assuming each sequence in the clusters table is a representative of a larger set, this table maps the representative to the larger set. 
                Search of targets is on larger set, instead.
    '''
    clusters=clusters.copy()
    if mapping is not None:
        clusters['id']=clusters['id'].map(mapping)
    if map_func is not None:
        clusters['id']=clusters['id'].apply(lambda x: map_func(x))
    if rep_map is not None:
        clusters=clusters.rename(columns={'id':'rep'}).merge(rep_map, on='rep')
    color=mpl.colors.to_rgba(color)
    color=color[:3]+(alpha,)
    out={'color':c_template.format(*[int(x*255) for x in color]), 'levels':levels, 'label':label}
    for level in levels:
        # out[level]=clusters[clusters['id'].isin(targets)][level].unique().tolist()
        # save table indicating which targets were in which cluster
        # out[level]=clusters[clusters['id'].isin(targets)][['id', level]].groupby(level)['id'].apply(list).reset_index().rename(columns={level:'id','id':'value'})
        out[level]=clusters[clusters['id'].isin(targets)][['id', level]].groupby(level)['id'].apply(list).to_dict()
    return out

def OutlineClustersDirect(clusters, level, color='red'):
    return {level:clusters, 'color':mpl.colors.to_rgba(color), 'levels':[level]}


import pandas as pd
from sklearn.cluster import MiniBatchKMeans, BisectingKMeans
import multiprocessing as mp
import numpy as np

def Kmeans(embeddings, n_clusters, outlabel):
    # convert embeddings dict to numpy ndarray
    ids=np.array(list(embeddings.keys()))
    X=np.array(list(embeddings.values()))

    # run kmeans, using minibatch kmeans for speed
    if type(n_clusters)==int:
        n_clusters=[n_clusters]

    clusters={}
    for k in n_clusters:
        kmeans = MiniBatchKMeans(n_clusters=k, random_state=0)
        kmeans.fit(X)
        clusters[k]=kmeans.labels_
    
    # make dataframe and write to file
    df=pd.DataFrame.from_dict(clusters)
    df['id']=ids
    df=df[['id']+n_clusters]
    out_name=f'{outlabel}_kmeans.csv'
    df=df.sort_values('id', key=lambda x: x.astype(int))
    df.to_csv(out_name, index=False)
    return {'kmeans': out_name}

def KMeansBisecting(embeddings, n_clusters, ids, out_dir='', out_prefix=''):

    # run kmeans, using minibatch kmeans for speed
    if type(n_clusters)==int:
        n_clusters=[n_clusters]

    # clusters={}
    for k in n_clusters:
        print(f'Running bisecting kmeans with {k} clusters')
        kmeans = BisectingKMeans(n_clusters=k, random_state=0)
        kmeans.fit(embeddings)
        # make a dataframe, with id and cluster assignments
        
        pd.DataFrame({'id':ids, 'cluster':kmeans.labels_}).to_csv(f'{out_dir}{out_prefix}_kmeans_bisecting_{k}.csv', index=False)

### Bisecting KMeans with tree structure
### This is a modified version of the BisectingKMeans class from sklearn
### The only change is that the tree structure is kept, so that the cluster assignments can be traced back to the original data

from sklearn.base import _fit_context
from sklearn.utils._openmp_helpers import _openmp_effective_n_threads
# from sklearn.utils._param_validation import Integral, Interval, StrOptions
from sklearn.utils.extmath import row_norms
from sklearn.utils.validation import _check_sample_weight, check_is_fitted, check_random_state, validate_data
from sklearn.cluster._k_means_common import _inertia_dense, _inertia_sparse
from sklearn.cluster._kmeans import (
    _BaseKMeans,
    _kmeans_single_elkan,
    _kmeans_single_lloyd,
    _labels_inertia_threadpool_limit,
)
from sklearn.cluster._bisect_k_means import _BisectingTree
import scipy.sparse as sp

class BisectingKMeansKeepTree(BisectingKMeans):

    # subclass of BisectingKMeans that keeps the tree structure
    def __init__(self, n_clusters=2, *, max_iter=100, tol=1e-4, random_state=None, verbose=0):
        super().__init__(n_clusters=n_clusters, max_iter=max_iter, tol=tol, random_state=random_state, verbose=verbose)
    
    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y=None, sample_weight=None):
        """Only change is to not erase the tree labels"""
        X = validate_data(
            self,
            X,
            accept_sparse="csr",
            dtype=[np.float64, np.float32],
            order="C",
            copy=self.copy_x,
            accept_large_sparse=False,
        )

        self._check_params_vs_input(X)

        self._random_state = check_random_state(self.random_state)
        sample_weight = _check_sample_weight(sample_weight, X, dtype=X.dtype)
        self._n_threads = _openmp_effective_n_threads()

        if self.algorithm == "lloyd" or self.n_clusters == 1:
            self._kmeans_single = _kmeans_single_lloyd
            self._check_mkl_vcomp(X, X.shape[0])
        else:
            self._kmeans_single = _kmeans_single_elkan

        # Subtract of mean of X for more accurate distance computations
        if not sp.issparse(X):
            self._X_mean = X.mean(axis=0)
            X -= self._X_mean

        # Initialize the hierarchical clusters tree
        self._bisecting_tree = _BisectingTree(
            indices=np.arange(X.shape[0]),
            center=X.mean(axis=0),
            score=0,
        )

        x_squared_norms = row_norms(X, squared=True)

        for _ in range(self.n_clusters - 1):
            # Chose cluster to bisect
            cluster_to_bisect = self._bisecting_tree.get_cluster_to_bisect()

            # Split this cluster into 2 subclusters
            self._bisect(X, x_squared_norms, sample_weight, cluster_to_bisect)

        # Aggregate final labels and centers from the bisecting tree
        self.labels_ = np.full(X.shape[0], -1, dtype=np.int32)
        self.cluster_centers_ = np.empty((self.n_clusters, X.shape[1]), dtype=X.dtype)

        for i, cluster_node in enumerate(self._bisecting_tree.iter_leaves()):
            self.labels_[cluster_node.indices] = i
            self.cluster_centers_[i] = cluster_node.center
            cluster_node.label = i  # label final clusters for future prediction
            # cluster_node.indices = None  # release memory # Only change

        # Restore original data
        if not sp.issparse(X):
            X += self._X_mean
            self.cluster_centers_ += self._X_mean

        _inertia = _inertia_sparse if sp.issparse(X) else _inertia_dense
        self.inertia_ = _inertia(
            X, sample_weight, self.cluster_centers_, self.labels_, self._n_threads
        )

        self._n_features_out = self.cluster_centers_.shape[0]

        return self

### Helpers for tree structure
import pandas as pd
# from sklearn.cluster import MiniBatchKMeans, BisectingKMeans
# import sklearn.cluster as cluster
# import importlib
# importlib.reload(cluster)
import numpy as np
# import pickle
# from Bio import Phylo
# import PCA
# from sklearn.decomposition import PCA

# make tree structure
def to_newick(node):
    if node is None:
        return ""
    
    # If the node is a leaf, return its name
    if node.left is None and node.right is None:
        label = int(node.label)+1
        return label
    
    # Recursively process left and right children
    left_newick = to_newick(node.left)
    right_newick = to_newick(node.right)
    
    left_dist = abs(node.score - node.left.score)
    right_dist = abs(node.score - node.right.score)

    # Combine the parts with parentheses
    if left_newick and right_newick:
        return f"({left_newick}:{left_dist},{right_newick}:{right_dist})"
    elif left_newick:
        return f"({left_newick}:{left_dist})"
    elif right_newick:
        return f"({right_newick}:{right_dist})"
    
    return ""

def tree_to_newick(root):
    newick_str = to_newick(root)
    if newick_str:
        return newick_str + ";"
    return ""

def get_leaf_index_mapping(tree):
    """
    Generate a mapping of leaf names to their indices and vice versa.
    """
    leaves = tree.get_terminals()
    leaf_names = [leaf.name for leaf in leaves]
    leaf_to_index = {name: i  for i, name in enumerate(leaf_names)}
    index_to_leaf = {i : name for i, name in enumerate(leaf_names)}
    return leaf_to_index, index_to_leaf

def get_merge_info(node, min_dists):
    
    # assumes input only has internal nodes, that each node is numbered by index, that tree traversal is post order

    ### (left for reference, very slow) get distance as minimum distance of leaves between the two nodes
    ### using min distance gives the single linkage
    # print('calculating distance...')
    # left_dist=min([tree.distance(node, leaf) for leaf in node.clades[0].get_terminals()])
    # right_dist=min([tree.distance(node, leaf) for leaf in node.clades[1].get_terminals()])

    left= node.clades[0]
    right= node.clades[1]

    # assumes distance traversal is leaves first
    left_dist=(0 if left.is_terminal() else min_dists[left.name])+left.branch_length
    right_dist=(0 if right.is_terminal() else min_dists[right.name])+right.branch_length

    min_dists[node.name]=min(left_dist, right_dist)

    distance=left_dist + right_dist

    n= len(node.get_terminals())

    # Append linkage information
    return [left.name, right.name, distance, n, node]

import time
from tqdm import tqdm
def compute_linkage_from_tree(tree):
    """
    Compute the linkage matrix directly from a tree with indexed leaves.
    """
    print('Setting up...')
    leaf_to_index, _ = get_leaf_index_mapping(tree)
    leaves = tree.get_terminals()
    num_leaves = len(leaves)
    linkage_list = []
    index_to_node={}

    
    clade_list=[]
    # traverse tree and gather all internal clades
    i=num_leaves
    # node_indices={} 
    for clade in tree.find_clades():
        if clade.is_terminal():
            # node_indices[clade]=leaf_to_index[clade.name]
            # rename to leaf index
            clade.name=leaf_to_index[clade.name]
            continue
        clade.name=i # give clade an arbitrary index
        i+=1
        # node_indices[clade]=clade.name
        index_to_node[clade.name]=clade
        clade_list.append(clade)
    # index_to_node={v:k for k, v in node_indices.items()}
    # invert clade list to start closest to leaves
    clade_list=clade_list[::-1]

    min_dists={}
    
    # tstart=time.time() # debug
    # print('traversing tree...')
    # traverse(tree.root)
    for clade in tqdm(clade_list, desc='traversing tree'):
        merge_info=get_merge_info(clade, min_dists)
        if merge_info is not None:
            linkage_list.append(merge_info)
    # print('time taken:', time.time()-tstart)
    
    # Convert to numpy array for scipy linkage function
    linkage_list=sorted(linkage_list, key=lambda x: x[2])
    nodes=[x[4] for x in linkage_list]
    linkage_list=[[x[0], x[1], x[2], x[3]] for x in linkage_list]
    Z = np.array(linkage_list)
    # sort by distance
    Z=Z[Z[:,2].argsort()]

    # relabel any node with n>num_leaves by order of appearance
    c=num_leaves
    ordered_node_index={}
    # first get ordered indices of nodes
    for node in nodes:
        ordered_node_index[node]=c
        c+=1
    # then replace internal node indices with new indices
    for row in Z:
        for i in [0, 1]:
            if row[i]>=num_leaves:
                row[i]=ordered_node_index[index_to_node[row[i]]]
    
    return Z, [leaf.name for leaf in tree.get_terminals()]

# def convert_linkage_to_clusters(Z, num_clusters):
#     """
#     Convert linkage matrix to clusters using fcluster.
#     """
#     cluster_labels = fcluster(Z, num_clusters, criterion='maxclust')
#     return cluster_labels



import math
from multiprocessing import Pool
import os
from types import SimpleNamespace
from .circle_collision import Circle, GetRadii, MakeLayout, GetLimits
import pandas as pd
import yaml

def ReadConfig(config_file):
    config_dict=yaml.safe_load(open(config_file))
    config=SimpleNamespace(**config_dict)
    return config

def CommentedConfigString(config:SimpleNamespace):
    cstr=yaml.safe_dump(config.__dict__)
    commented='#'+cstr.replace('\n','\n#')
    return commented

def ConfigFromComments(text):
    cstr=text[1:].replace('\n#','\n')
    config=SimpleNamespace(**yaml.safe_load(cstr))
    return config

def process_child_clusters(args):
    '''Process child clusters in parallel
        Has same inputs as MakeLayout, but with child_clusters instead of radii as the first argument
    '''
    child_clusters, radii, radius_pad, space_scale, seed, seed_pos, cycles, parent, pull, expand_mult, velocity_iterations, position_iterations, time_step, metadata = args
    return MakeLayout(radii.loc[child_clusters], radius_pad=0, space_mult=space_scale*.8, seed=seed, seed_coords=seed_pos,
                      time_step=time_step, cycles=cycles, parent=parent, pull=pull, expand_mult=expand_mult,
                      velocity_iterations=velocity_iterations, position_iterations=position_iterations, metadata=metadata)

def CenterOfMass(circles): # not used
    x, y=0, 0
    for c in circles:
        x+=c.x
        y+=c.y
    x/=len(circles)
    y/=len(circles)
    return x, y

def Box2DLayout(clusters, levels, minsize=0, space_scale=1.1, pull=1, level_pull=None, radius_pad=3, force_inclusion=None, velocity_iterations=16, position_iterations=6,
                cycle_scale=1000, min_cycles=1000, max_cycles=100_000, time_step=0.1, level_time_step=None, pull_base=False, seed=None, seed_coords=None, expand_mult=1):
    
    '''
    Function to turn a table of cluster definitions into a hierarchical layout of circles

    Args:
        General:
            - clusters (pd.DataFrame): table of cluster definitions, with columns 'id' for each sequence one column per "level" of clustering
            - levels (list): list of column names in clusters to use as levels of clustering. Must be in ascending order (lowest cut-off first) of hierarchy
            - minsize (int): minimum size of a cluster to be included in the layout
            - force_inclusion (list): list of sequence ids to force inclusion in the layout
        Tuning layout:
            - pull (float): multiplier for the pull between circles when packing
            - level_pull (dict): dictionary of level-specific pull multipliers
            - radius_pad (float): padding between circles in the base (lowest) level
            - seed (int): random seed for layout. Use same seed to reproduce layout, assuming same input data and cycles of simulation
            - seed_coords (pd.DataFrame): table of seed coordinates for each cluster in the base level. Must have columns 'id', 'x', 'y'
            - pull_base (bool): whether to allow circles in the base level to pull each other. Still respects radius_pad, just makes the base level more compact.
            Niche:
                - space_scale (float): multiplier for the space between circles on initial randomized layout.
                - expand_mult (float): multiplier for the expansion of circles. Somewhat unpredictable, can increase speed if too many circles start overlapping, but quite situational.
        Box2D simulation:
            - velocity_iterations (int): number of velocity iterations in the Box2D simulation
            - position_iterations (int): number of position iterations in the Box2D simulation
            - cycle_scale (int): multiplier for the number of cycles in the Box2D simulation per 50 child clusters
            - max_cycles (int): maximum number of cycles in the Box2D simulation
            - time_step (int): time step in the Box2D simulation
            - level_time_step (dict): dictionary of level-specific time steps

    Returns:
        dict: dictionary of circles at each level of clustering. Each level is a dictionary of circles, with the id of the circle as the key.
              Each circle object contains its xy position, radius, and size (number of sequences in the cluster)
    '''
    
    ### Process child levels in parallel
    
    size_dicts={}
    if len(levels)>1:
        process_args = []
        for lv, level_id in enumerate(levels[1:], start=1):
            df=clusters.loc[:,['id', level_id]]
            size=df.groupby(level_id).size().rename('size')#.reset_index()
            size_dicts[level_id]=dict(size)
            forced=size.index.isin(df[df['id'].isin(force_inclusion)][level_id].unique()) if force_inclusion is not None else None
            
            radii=GetRadii(size, minsize, 1, forced)
            print(f'Level {level_id}, clusters: {len(radii)}')
            seed_pos=None
            if seed_coords is not None:
                seed_pos=seed_coords.merge(df, on='id')
                seed_pos=seed_pos[seed_pos[level_id].isin(radii.index)]
                seed_pos=seed_pos.groupby(level_id)[['x','y']].mean()

            parent_level=levels[lv-1]
            # subset=clusters[clusters[level_id].isin(radii.index) & clusters[levels[lv-1]].isin(data[parent_level].keys())]
            subset=clusters[clusters[level_id].isin(radii.index)]
            # print(subset)
            
            print('preparing child clusters')
            for cluster, group in subset.groupby(parent_level):
                child_clusters = group[level_id].unique()
                # print(f'Level {lv} ({level_id}), parent: {cluster}, clusters: {len(group)}, children: {len(child_clusters)}')
                cycles=min( math.ceil(max(len(child_clusters)/50, 1) * cycle_scale), max_cycles )
                cycles=max(cycles, min_cycles)

                if level_pull is not None and level_id in level_pull:
                    cur_pull=level_pull[level_id]
                    # print(f'Adjusting pull for level {level_id} to {cur_pull}')
                else:
                    cur_pull=pull
                
                if level_time_step is not None and level_id in level_time_step:
                    cur_time_step=level_time_step[level_id]
                    # print(f'Adjusting time step for level {level_id} to {cur_time_step}')
                else:
                    cur_time_step=time_step
                process_args.append((child_clusters, radii, radius_pad, space_scale, seed, seed_pos,
                                        cycles, None, cur_pull, expand_mult,
                                        velocity_iterations, position_iterations, cur_time_step, {'plevel':parent_level, 'parent':cluster,'level':level_id}))    

        print('starting pool')
        with Pool() as pool:
            results = list(pool.imap_unordered(process_child_clusters, process_args))
    else:
        results=[]

    ### Process base level to avoid collisions
    level_id=levels[0]
    df=clusters.loc[:,['id', level_id]]
    size=df.groupby(level_id).size().rename('size')#.reset_index()
    size_dicts[level_id]=dict(size)
    forced=size.index.isin(df[df['id'].isin(force_inclusion)][level_id].unique()) if force_inclusion is not None else None
    
    radii=GetRadii(size, minsize, 1, forced)
    radius_override={}
    if len(levels)>1:
        child_level=levels[1]
        children={}
        for circles, meta in results:
            if meta['level']==child_level:
                children[meta['parent']]=circles

        for cluster, radius in radii.items():
            if cluster not in children:
                continue
            child_circles=children[cluster]
            # get center of mass
            # x, y=CenterOfMass(child_circles)
            x0, x1, y0, y1=GetLimits(child_circles)
            cx=(x0+x1)/2
            cy=(y0+y1)/2
            max_dist=0
            # distance to each child circle + radius
            for c in child_circles:
                dist=((c.x-cx)**2+(c.y-cy)**2)**0.5 + c.r
                if dist>max_dist:
                    max_dist=dist
            if max_dist>radius:
                radii.loc[cluster]=max_dist
                radius_override[cluster]=radius

    print(f'Level {level_id}, clusters: {len(radii)}')
    seed_pos=None
    if seed_coords is not None:
        seed_pos=seed_coords.merge(df, on='id')
        seed_pos=seed_pos[seed_pos[level_id].isin(radii.index)]
        seed_pos=seed_pos.groupby(level_id)[['x','y']].mean()
    base_circles, _ =MakeLayout(radii, radius_pad=radius_pad, space_mult=space_scale, seed=seed, 
                               seed_coords=seed_pos, pull=pull if pull_base else 0, expand_mult=expand_mult,
                               velocity_iterations=velocity_iterations, position_iterations=position_iterations, time_step=time_step)
    
    for c in base_circles:
        if c.id in radius_override:
            c.r=radius_override[c.id]
        c.size=size_dicts[level_id][c.id]
    # collapse all data
    data={}
    data[level_id]=dict((c.id, c) for c in base_circles)
    grouped={}
    for circles, meta in results:
        for c in circles:
            c.size=size_dicts[meta['level']][c.id]
        if meta['level'] not in data:
            grouped[meta['level']]={meta['parent']:{'plevel':meta['plevel'], 'data':circles}}
            data[meta['level']]=dict((c.id, c) for c in circles)
        else:
            grouped[meta['level']][meta['parent']]={'plevel':meta['plevel'], 'data':circles}
            data[meta['level']].update((c.id, c) for c in circles)
    # for each level, recenter the circles into the parent circle
    for lv, level_id in enumerate(levels[1:], start=1):
        for parent, info in grouped[level_id].items():
            parent_circle=data[info['plevel']][parent]
            child_circles=info['data']
            x0, x1, y0, y1=GetLimits(child_circles)
            cx=(x0+x1)/2
            cy=(y0+y1)/2

            px, py=parent_circle.x, parent_circle.y
            for i, c in enumerate(child_circles):
                c.x=c.x-cx+px
                c.y=c.y-cy+py
    return data

def SaveLayout(data, outfile, config:SimpleNamespace=None):
    rows=[]
    for lv, c_dict in data.items():
        for c in c_dict.values():
            rows.append([lv, c.id, c.x, c.y, c.r])
    tb=pd.DataFrame(rows, columns=['level','cluster','x','y','r'])
    # print(tb)
    with open(outfile, 'w') as f:
        if config is not None:
            print(CommentedConfigString(config), file=f)
    tb.to_csv(outfile, mode='a', index=False)

def ReadLayout(file, config_check:SimpleNamespace=None):
    if not os.path.exists(file):
        return None

    if config_check is not None:
        saved_cstr=''
        with open(file) as f:
            for line in f:
                if not line.startswith('#'):
                    break
                saved_cstr+=line
        saved_config=ConfigFromComments(saved_cstr)
        # print(saved_config, config_check)
        if saved_config!=config_check:
            return None

    # re-use old layout
    tb=pd.read_csv(file, comment='#')
    data={}
    for lv, group in tb.groupby('level'):
        data[lv]={}
        for row in group.itertuples():
            data[lv][row.cluster]=Circle(row.x,row.y,row.r,row.cluster)
    return data



import pickle
from scipy.cluster.hierarchy import fcluster, to_tree#, linkage
from fastcluster import linkage
from Bio import Phylo
import numpy as np
import pandas as pd
import os

def HierarchicalClustering(ids, embeddings, out_dir='', out_prefix='', method='weighted', metric='cosine'):

    # run hierarchical clustering
    Z = linkage(embeddings, method=method, metric=metric)

    # save linkage matrix
    outfile=f'{out_prefix}_linkage_matrix.pkl'
    idsfile=f'{out_prefix}_linkage_ids.pkl'
    with open(os.path.join(out_dir, outfile), 'wb') as f:
        pickle.dump(Z, f)
    with open(os.path.join(out_dir, idsfile), 'wb') as f:
        pickle.dump(ids, f)
    return {'linkage_matrix': outfile, 'linkage_ids': idsfile}

def FlattenHierarchy(linkage_matrix, linkage_ids, cutoffs, out_dir='', out_prefix='', criterion='maxclust'):
    # load linkage matrix
    with open(linkage_matrix, 'rb') as f:
        Z = pickle.load(f)
    # load ids
    with open(linkage_ids, 'rb') as f:
        ids = pickle.load(f)

    # flatten hierarchy
    for cutoff in cutoffs:
        clu=fcluster(Z, cutoff, criterion=criterion)
        # make dataframe and write to file
        df=pd.DataFrame({'id':ids, 'cluster':clu})
        df.to_csv(os.path.join(out_dir, f'{out_prefix}_hierarchical_flat_{cutoff}.csv'), index=False)

def linkage_to_newick(linkage_matrix, leaf_names):
    def get_newick(node, newick, parentdist, leaf_names):
        if node.is_leaf():
            return f"{leaf_names[node.id]}:{parentdist - node.dist}{newick}"
        else:
            if len(newick) > 0:
                newick = f"):{node.dist}{newick}"
            else:
                newick = ");"
            newick = get_newick(node.get_left(), newick, node.dist, leaf_names)
            newick = get_newick(node.get_right(), ",%s" % (newick), node.dist, leaf_names)
            newick = "(%s" % (newick)
            return newick

    tree = to_tree(linkage_matrix, False)
    newick = get_newick(tree, "", tree.dist, leaf_names)
    return newick

def NameInternalNodes(treefile):
    tree=Phylo.read(treefile, 'newick')
    # travel through tree and name internal nodes
    node=0
    for clade in tree.get_nonterminals(order='preorder'):
        
        clade.name=f'node_{node}'
        node+=1
    write_newick(tree, treefile)

def write_newick(tree, out_file):
    def _write_node(node):
        node_name = f"{node.name}:{node.branch_length}"
        if node.is_terminal():
            return node_name
        else:
            children = ','.join(_write_node(child) for child in node.clades)
            return f"({children}){node_name}"

    newick_string = _write_node(tree.root) + ";"
    with open(out_file, "w") as file_handle:
        file_handle.write(newick_string + "\n")


import matplotlib.pyplot as plt

from layout.circle_collision import GetLimits

def DrawCircles(layout, out_label):
    '''Debug function to draw a layout of circles to a file'''
    all_c=[]
    for lv, c_dict in layout.items():
        all_c+=list(c_dict.values())
    xmin, xmax, ymin, ymax=GetLimits(all_c, 5)

    f, ax=plt.subplots(figsize=(10,10))
    ax.axis('off')
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    ax.set_aspect('equal', adjustable='box')
    for lv, c_dict in layout.items():
        lv_frac=lv/len(layout)
        for c in c_dict.values():
            # draw circle
            circle=plt.Circle((c.x, c.y), c.r, color='b', lw=.5, alpha=.2+lv_frac*.8, fill=False)
            ax.add_patch(circle)
    f.savefig(f'{out_label}_layout.png', dpi=300, bbox_inches='tight')